import datetime
import logging
from collections import defaultdict

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import F, Max, Min
from django.utils import timezone
from django.utils.formats import localize
from django.utils.translation import gettext_lazy as _, ngettext_lazy

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo

from .. import app_settings, providers
from .assets import CharacterAsset
from .audits import CharacterLocation, EveLocation
from .clones import Clone, JumpClone
from .eve_models import (
    EveItemCategory, EveItemGroup, EveItemType, MapConstellation, MapRegion,
    MapSystem,
)
from .interactions import CharacterTitle, CorporationHistory
from .skills import SkillList, SkillTotals

logger = logging.getLogger(__name__)


class FilterBase(models.Model):

    name = models.CharField(max_length=500)
    description = models.CharField(max_length=500)

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.name}: {self.description}"

    def process_filter(self, user: User):
        raise NotImplementedError("Please Create a filter!")


class FullyLoadedFilter(FilterBase):
    reversed_logic = models.BooleanField(
        default=False, help_text="If set all members WITHOUT audit fully loaded will pass the test.")
    count_message_only = models.BooleanField(default=False)

    class Meta:
        verbose_name = "Smart Filter: Audit Fully Loaded"
        verbose_name_plural = verbose_name

    def process_filter(self, user: User):
        logic = self.reversed_logic
        try:
            character_list = user.character_ownerships.all() \
                .select_related('character', 'character__characteraudit')

            character_count = character_list.filter(
                character__characteraudit__isnull=True).count()
            if character_count == 0:
                valid_audits = 0
                character_cnt = 0
                for c in character_list:
                    if c.character.characteraudit.is_active():
                        valid_audits += 1
                    character_cnt += 1
                if valid_audits == character_cnt:
                    return not logic
                else:
                    return logic
            else:
                return logic
        except Exception as e:
            logger.error(e, exc_info=1)
            return False

    def process_field(self, users):
        return self.audit_filter(users)

    def audit_filter(self, users):
        logic = self.reversed_logic
        co = CharacterOwnership.objects.filter(user__in=users).select_related(
            'user', 'character__characteraudit')
        chars = {}
        for c in co:
            if c.user.id not in chars:
                chars[c.user.id] = []
            try:
                if not c.character.characteraudit.is_active():
                    chars[c.user.id].append(c.character.character_name)
            except ObjectDoesNotExist:
                chars[c.user.id].append(c.character.character_name)

        output = defaultdict(lambda: {"message": "", "check": logic})
        for u in users:
            c = chars.get(u.id, False)
            if c is not False:
                if len(c) > 0:
                    _msg = ", ".join(c)
                    if self.count_message_only:
                        _msg = len(c)
                    output[u.id] = {"message": _msg, "check": logic}
                    continue
                else:
                    output[u.id] = {
                        "message": "All Characters Loaded", "check": not logic}
                    continue
            output[u.id] = {"message": "", "check": logic}
        return output


class HighestSPFilter(FilterBase):
    sp_cutoff = models.BigIntegerField(default=5000000000)
    swap_logic = models.BooleanField(default=False)

    class Meta:
        verbose_name = "Smart Filter: Highest SP Character"
        verbose_name_plural = f"{verbose_name}"

    def process_filter(self, user: User):
        try:
            return self.audit_filter([user])[user.id]['check']
        except Exception as e:
            logger.error(e, exc_info=1)
            return False

    def audit_filter(self, users):
        co = CharacterOwnership.objects.filter(
            user__in=users).select_related('user', 'character')
        chars = {}
        all_sp = SkillTotals.objects.filter(
            character__character__in=co.values_list('character'))
    # total_sp = models.BigIntegerField()
    # unallocated_sp = models.IntegerField(null=True, default=None)

        failure = self.swap_logic
        for i in all_sp:
            uid = i.character.character.character_ownership.user.id
            tsp = i.total_sp
            if i.unallocated_sp:
                tsp += i.unallocated_sp

            if uid not in chars:
                chars[uid] = tsp
            elif tsp > chars[uid]:
                chars[uid] = tsp

        output = defaultdict(lambda: {"message": 0, "check": False})
        for u in users:
            c = chars.get(u.id, 0)
            if c < self.sp_cutoff:
                output[u.id] = {"message": c, "check": failure}
                continue
            else:
                output[u.id] = {"message": c, "check": not failure}
                continue
        return output


class TimeInCorpFilter(FilterBase):
    class Meta:
        verbose_name = "Smart Filter: Main's Time in Corp"
        verbose_name_plural = verbose_name

    days_in_corp = models.IntegerField(default=30)

    reversed_logic = models.BooleanField(
        default=False,
        help_text="If set all members less than the days in corp will pass the test."
    )

    def process_filter(self, user: User):
        logic = self.reversed_logic
        try:
            main_character = user.profile.main_character.characteraudit
            histories = CorporationHistory.objects.filter(
                character=main_character
            ).order_by('-start_date').first()

            days = timezone.now() - histories.start_date

            if main_character.character.corporation_id != histories.corporation_id:
                return False  # FAIL if history is no god from CCP.

            if days.days >= self.days_in_corp:
                return not logic
            else:
                return logic
        except Exception as e:
            # logger.error(e, exc_info=1)
            return False

    def audit_filter(self, users):
        logic = self.reversed_logic
        histories = CorporationHistory.objects.filter(
            character__character_id__in=users.values_list(
                "profile__main_character", flat=True)
        ).values(uid=F("character__character__character_ownership__user_id")).annotate(
            max_id=Max("record_id"),
        )
        # pull the specific histories
        histories = CorporationHistory.objects.filter(
            record_id__in=histories.values_list("max_id", flat=True)
        ).values(
            ccid=F("character__character__corporation_id"),
            rcid=F("corporation_id"),
            uid=F("character__character__character_ownership__user_id"),
            std=F("start_date")
        )

        chars = defaultdict(lambda: {})
        for c in histories:
            if c["ccid"] != c["rcid"]:
                continue  # Skip if not main corp.

            if c['std']:
                days = timezone.now() - c['std']
                days = days.days
            else:
                days = -1
            chars[c['uid']] = {
                "start_date": c["std"],
                "days_in_corp": days
            }

        output = defaultdict(lambda: {"message": "", "check": False})

        for c, char_corp_data in chars.items():
            start_date_localized = localize(char_corp_data["start_date"].date(
            )) if char_corp_data["start_date"] is not None else None
            end_date_localized = None

            if char_corp_data["days_in_corp"] >= self.days_in_corp:
                check = not logic
            else:
                if start_date_localized:
                    end_date_localized = localize(
                        (
                            char_corp_data["start_date"] +
                            datetime.timedelta(days=self.days_in_corp)
                        ).date()
                    )

                check = logic
            if char_corp_data["days_in_corp"] < 0:
                msg = "No Audit"
                check = False
            else:
                msg = (
                    ngettext_lazy(
                        singular=f'{char_corp_data["days_in_corp"]:d} day (Since: {start_date_localized})',
                        plural=f'{char_corp_data["days_in_corp"]:d} days (Since: {start_date_localized})',
                        number=char_corp_data["days_in_corp"],
                    )
                    if not end_date_localized
                    else ngettext_lazy(
                        singular=f'{char_corp_data["days_in_corp"]:d} day (Until: {end_date_localized})',
                        plural=f'{char_corp_data["days_in_corp"]:d} days (Until: {end_date_localized})',
                        number=char_corp_data["days_in_corp"],
                    )
                )
            output[c] = {"message": msg, "check": check}
        return output


class AssetsFilter(FilterBase):
    class Meta:
        verbose_name = "Smart Filter: Assets in Locations"
        verbose_name_plural = verbose_name
    count_message_only = models.BooleanField(default=False)

    types = models.ManyToManyField(EveItemType, blank=True,
                                   help_text="Filter on Asset Types.")
    groups = models.ManyToManyField(EveItemGroup, blank=True,
                                    help_text="Filter on Asset Groups.")
    categories = models.ManyToManyField(EveItemCategory, blank=True,
                                        help_text="Filter on Asset Categories.")

    systems = models.ManyToManyField(MapSystem, blank=True,
                                     help_text="Limit filter to specific systems")
    constellations = models.ManyToManyField(MapConstellation, blank=True,
                                            help_text="Limit filter to specific constellations")
    regions = models.ManyToManyField(MapRegion, blank=True,
                                     help_text="Limit filter to specific regions")

    reversed_logic = models.BooleanField(
        default=False, help_text="Negate the value of the filter, i.e. check for absence of assets")

    def filter_query(self, users):
        character_list = CharacterOwnership.objects.filter(user__in=users) \
            .select_related('character', 'character__characteraudit')
        cnt_types = self.types.all().count()
        cnt_groups = self.groups.all().count()
        cnt_cats = self.categories.all().count()

        character_count = CharacterAsset.objects.filter(
            character__character__in=character_list.values_list('character'))
        output = []

        if cnt_types > 0:
            output.append(models.Q(type_name__in=self.types.all()))

        if cnt_groups > 0:
            output.append(models.Q(type_name__group__in=self.groups.all()))

        if cnt_cats > 0:
            output.append(
                models.Q(type_name__group__category__in=self.categories.all()))

        if len(output) == 0:
            return False

        query = output.pop()
        for _q in output:
            query |= _q
        character_count = character_count.filter(query)

        output = []

        if self.systems.all().count() > 0:
            output.append(
                models.Q(location_name__system__in=self.systems.all()))
            output.append(models.Q(
                location_id__in=self.systems.all().values_list('system_id', flat=True)))
        if self.constellations.all().count() > 0:
            output.append(
                models.Q(location_name__system__constellation__in=self.constellations.all()))
            output.append(models.Q(location_id__in=MapSystem.objects.filter(
                constellation__in=self.constellations.all()).values_list('system_id', flat=True)))
        if self.regions.all().count() > 0:
            output.append(
                models.Q(location_name__system__constellation__region__in=self.regions.all()))
            output.append(models.Q(location_id__in=MapSystem.objects.filter(
                constellation__region__in=self.regions.all()).values_list('system_id', flat=True)))

        if len(output) > 0:
            query = output.pop()
            for _q in output:
                query |= _q
            character_count = character_count.filter(query)
        return character_count

    def process_filter(self, user: User):
        try:
            co = self.filter_query([user])
            # XOR with self.reversed_logic
            return (co.count() > 0) != self.reversed_logic
        except Exception as e:
            logger.error(e, exc_info=1)
            return False != self.reversed_logic

    def audit_filter(self, users):
        co = self.filter_query(users).values("character__character__character_ownership__user",
                                             "type_name__name", "character__character__character_name")
        chars = defaultdict(dict)
        for c in co:
            uid = c["character__character__character_ownership__user"]
            char_name = c["character__character__character_name"]
            asset_type = c['type_name__name']
            if char_name not in chars[uid]:
                chars[uid][char_name] = []
            chars[uid][char_name].append(asset_type)

        output = defaultdict(
            lambda: {"message": "", "check": False != self.reversed_logic})
        for u in users:
            if len(chars[u.id]) > 0:
                dread_count = 0
                out_message = []
                for char, char_items in chars[u.id].items():
                    dread_count += len(char_items)
                    out_message.append(
                        f"{char}: {', '.join(list(set(char_items)))}")
                if self.count_message_only:
                    output[u.id] = {"message": dread_count,
                                    "check": True != self.reversed_logic}
                else:
                    output[u.id] = {"message": "<br>".join(
                        out_message), "check": True != self.reversed_logic}
            else:
                output[u.id] = {"message": "",
                                "check": False != self.reversed_logic}
        return output


class CurrentShipFilter(FilterBase):
    class Meta:
        verbose_name = "Smart Filter: The Current Ship being Flown"
        verbose_name_plural = verbose_name
    count_message_only = models.BooleanField(default=False)

    types = models.ManyToManyField(EveItemType, blank=True,
                                   help_text="Filter on Ship Types.")
    groups = models.ManyToManyField(EveItemGroup, blank=True,
                                    help_text="Filter on Ship Groups.")

    systems = models.ManyToManyField(MapSystem, blank=True,
                                     help_text="Limit filter to specific systems")
    constellations = models.ManyToManyField(MapConstellation, blank=True,
                                            help_text="Limit filter to specific constellations")
    regions = models.ManyToManyField(MapRegion, blank=True,
                                     help_text="Limit filter to specific regions")

    def filter_query(self, users):
        character_list = CharacterOwnership.objects.filter(user__in=users) \
            .select_related('character', "character__characteraudit")
        cnt_types = self.types.all().count()
        cnt_groups = self.groups.all().count()

        character_count = CharacterLocation.objects.filter(
            character__character__in=character_list.values_list('character'))
        output = []

        if cnt_types > 0:
            output.append(models.Q(current_ship__in=self.types.all()))

        if cnt_groups > 0:
            output.append(models.Q(current_ship__group__in=self.groups.all()))

        if len(output) == 0:
            return False

        query = output.pop()
        for _q in output:
            query |= _q
        character_count = character_count.filter(query)

        output = []

        if self.systems.all().count() > 0:
            output.append(
                models.Q(
                    current_location__system__in=self.systems.all()
                )
            )
            output.append(
                models.Q(
                    current_location__location_id__in=self.systems.all(
                    ).values_list('system_id', flat=True)
                )
            )
        if self.constellations.all().count() > 0:
            output.append(
                models.Q(
                    current_location__system__constellation__in=self.constellations.all()
                )
            )
            output.append(
                models.Q(
                    current_location__location_id__in=MapSystem.objects.filter(
                        constellation__in=self.constellations.all()
                    ).values_list('system_id', flat=True)
                )
            )
        if self.regions.all().count() > 0:
            output.append(
                models.Q(
                    current_location__system__constellation__region__in=self.regions.all()
                )
            )
            output.append(
                models.Q(
                    current_location__location_id__in=MapSystem.objects.filter(
                        constellation__region__in=self.regions.all()
                    ).values_list('system_id', flat=True)
                )
            )

        if len(output) > 0:
            query = output.pop()
            for _q in output:
                query |= _q
            character_count = character_count.filter(query)
        return character_count

    def process_filter(self, user: User):
        try:
            co = self.filter_query([user])
            if co.count() > 0:
                return True
            else:
                return False
        except Exception as e:
            logger.error(e, exc_info=1)
            return False

    def audit_filter(self, users):
        co = self.filter_query(
            users
        ).values(
            "character__character__character_ownership__user",
            "character__character__character_name",
            "current_ship_name"
        )
        chars = defaultdict(dict)
        for c in co:
            uid = c["character__character__character_ownership__user"]
            char_name = c["character__character__character_name"]
            ship_type = c['current_ship_name']
            if char_name not in chars[uid]:
                chars[uid][char_name] = []
            chars[uid][char_name].append(ship_type)

        output = defaultdict(lambda: {"message": "", "check": False})
        for u in users:
            if len(chars[u.id]) > 0:
                ship_count = 0
                out_message = []
                for char, char_items in chars[u.id].items():
                    ship_count += len(char_items)
                    out_message.append(
                        f"{char}: {', '.join(list(set(char_items)))}")
                if self.count_message_only:
                    output[u.id] = {"message": ship_count, "check": True}
                else:
                    output[u.id] = {"message": "<br>".join(
                        out_message), "check": True}
            else:
                output[u.id] = {"message": "", "check": False}
        return output


class Skillfilter(FilterBase):
    class Meta:
        verbose_name = "Smart Filter: Skill list checks"
        verbose_name_plural = verbose_name

    count_message_only = models.BooleanField(default=False)

    required_skill_lists = models.ManyToManyField(SkillList, blank=True)
    single_req_skill_lists = models.ManyToManyField(
        SkillList, blank=True, related_name="single_req")

    def process_filter(self, user: User):
        try:  # avatar 11567
            skills_list = providers.skills.get_and_cache_user(user.id)
            skill_lists = self.required_skill_lists.all()
            req_one = self.single_req_skill_lists.all()
            if skill_lists.count() == 0 and req_one.count() == 0:
                return False

            skill_list_base = {}
            for skl in skill_lists:
                skill_list_base[skl.name] = {}
                skill_list_base[skl.name]['pass'] = False

            if req_one.count() > 0:
                skill_list_single = {}
                for skl in req_one:
                    skill_list_single[skl.name] = {}
                    skill_list_single[skl.name]['pass'] = False

            skill_tables = skills_list.get("skills_list")

            for char in skill_tables:
                for d_name, d_list in skill_list_base.items():
                    if len(skill_tables[char]["doctrines"][d_name]) == 1:
                        skill_list_base[d_name]['pass'] = True
            if req_one.count() > 0:
                single_pass = False
                for char in skill_tables:
                    for d_name, d_list in skill_list_single.items():
                        if len(skill_tables[char]["doctrines"][d_name]) == 1:
                            single_pass = True
                            break

            result = True
            for skill_list, skills_result in skill_list_base.items():
                result = result and skills_result['pass']
            if req_one.count() > 0:
                return result and single_pass
            else:
                return result

        except Exception as e:
            logger.error(e, exc_info=1)
            return False

    def audit_filter(self, users):
        output = defaultdict(lambda: {"message": "No Data", "check": False})
        accounts = providers.skills.get_and_cache_users(users)
        for uid, u in accounts.items():
            message = []
            skill_lists = self.required_skill_lists.all()
            req_one = self.single_req_skill_lists.all()
            if skill_lists.count() == 0 and req_one.count() == 0:
                return False

            skill_list_base = {}
            for skl in skill_lists:
                skill_list_base[skl.name] = {}
                skill_list_base[skl.name]['pass'] = False

            if req_one.count() > 0:
                skill_list_single = {}
                for skl in req_one:
                    skill_list_single[skl.name] = {}
                    skill_list_single[skl.name]['pass'] = False

            try:
                skill_tables = u['data'].get("skills_list")

                for char in skill_tables:
                    for d_name, d_list in skill_list_base.items():
                        if len(skill_tables[char]["doctrines"][d_name]) == 1:
                            skill_list_base[d_name]['pass'] = True
                            message.append(f"{char}: {d_name}")

                if req_one.count() > 0:
                    single_pass = False
                    for char in skill_tables:
                        for d_name, d_list in skill_list_single.items():
                            if len(skill_tables[char]["doctrines"][d_name]) == 1:
                                single_pass = True
                                message.append(f"{char}: {d_name}")
                                break

                result = True
                for skill_list, skills_result in skill_list_base.items():
                    result = result and skills_result['pass']
                _msg = "<br>".join(message)
                if self.count_message_only:
                    _msg = len(message)
                if req_one.count() > 0:
                    output[uid] = {'check': result and single_pass,
                                   'message': _msg}
                else:
                    output[uid] = {'check': result,
                                   'message': _msg}
            except KeyError:
                pass

        return output


class Rolefilter(FilterBase):
    class Meta:
        verbose_name = "Smart Filter: Corporate Role checks"
        verbose_name_plural = verbose_name

    has_director = models.BooleanField(default=False)
    has_accountant = models.BooleanField(default=False)
    has_station_manager = models.BooleanField(default=False)
    has_personnel_manager = models.BooleanField(default=False)

    main_only = models.BooleanField(default=False)

    corps_filter = models.ManyToManyField(
        EveCorporationInfo, related_name='audit_role_filters', blank=True)

    alliances_filter = models.ManyToManyField(
        EveAllianceInfo, related_name='audit_role_filters', blank=True)

    def process_filter(self, user: User):
        try:
            characters = user.character_ownerships.all()
            queries = []
            if self.main_only:
                characters = characters.filter(
                    character__character_id=user.profile.main_character.character_id)
            if self.corps_filter.all().count():
                characters = characters.filter(
                    character__corporation_id__in=self.corps_filter.all().values_list("corporation_id", flat=True))
            if self.alliances_filter.all().count():
                characters = characters.filter(
                    character__alliance_id__in=self.alliances_filter.all().values_list("alliance_id", flat=True))
            if self.has_director:
                _q = models.Q(
                    character__characteraudit__characterroles__director=True)
                queries.append(_q)
            if self.has_accountant:
                _q = models.Q(
                    character__characteraudit__characterroles__accountant=True)
                queries.append(_q)
            if self.has_station_manager:
                _q = models.Q(
                    character__characteraudit__characterroles__station_manager=True)
                queries.append(_q)
            if self.has_personnel_manager:
                _q = models.Q(
                    character__characteraudit__characterroles__personnel_manager=True)
                queries.append(_q)
            query = queries.pop()
            for q in queries:
                query |= q
            if characters.filter(query).exists():
                return True
            else:
                return False
        except Exception as e:
            logger.error(e, exc_info=1)
            return False

    def audit_filter(self, users):

        co = CharacterOwnership.objects.filter(user__in=users)
        queries = []
        if self.main_only:
            co = co.filter(character__character_id=models.F(
                "user__profile__main_character__character_id"))
        if self.corps_filter.all().count():
            co = co.filter(
                character__corporation_id__in=self.corps_filter.all().values_list("corporation_id", flat=True))
        if self.alliances_filter.all().count():
            co = co.filter(
                character__alliance_id__in=self.alliances_filter.all().values_list("alliance_id", flat=True))
        if self.has_director:
            _q = models.Q(
                character__characteraudit__characterroles__director=True)
            queries.append(_q)
        if self.has_accountant:
            _q = models.Q(
                character__characteraudit__characterroles__accountant=True)
            queries.append(_q)
        if self.has_station_manager:
            _q = models.Q(
                character__characteraudit__characterroles__station_manager=True)
            queries.append(_q)
        if self.has_personnel_manager:
            _q = models.Q(
                character__characteraudit__characterroles__personnel_manager=True)
            queries.append(_q)
        query = queries.pop()
        for q in queries:
            query |= q

        co = co.filter(query)

        chars = {}
        for c in co:
            if c.user.id not in chars:
                chars[c.user.id] = []
            chars[c.user.id].append(c.character.character_name)

        output = defaultdict(lambda: {"message": "", "check": False})
        for u in users:
            c = chars.get(u.id, False)
            if c is not False:
                if len(c) > 0:
                    output[u.id] = {"message": ", ".join(c), "check": True}
                    continue
                else:
                    output[u.id] = {"message": "", "check": False}
                    continue
            output[u.id] = {"message": "", "check": False}
        return output


class Titlefilter(FilterBase):
    class Meta:
        verbose_name = "Smart Filter: Corporate Title checks"
        verbose_name_plural = verbose_name

    titles = models.ForeignKey(
        CharacterTitle, on_delete=models.CASCADE)

    def process_filter(self, user: User):
        try:
            characters = user.character_ownerships.all()

            if characters.filter(character__characteraudit__characterroles__titles__in=[self.titles]).exists():
                return True
            else:
                return False
        except Exception as e:
            logger.error(e, exc_info=1)
            return False

    def audit_filter(self, users):
        co = CharacterOwnership.objects.filter(user__in=users,
                                               character__characteraudit__characterroles__titles__in=[self.titles])

        chars = {}
        for c in co:
            if c.user.id not in chars:
                chars[c.user.id] = []
            chars[c.user.id].append(c.character.character_name)

        output = defaultdict(lambda: {"message": "", "check": False})
        for u in users:
            c = chars.get(u.id, False)
            if c is not False:
                if len(c) > 0:
                    output[u.id] = {"message": ", ".join(c), "check": True}
                    continue
                else:
                    output[u.id] = {"message": "", "check": False}
                    continue
            output[u.id] = {"message": "", "check": False}
        return output


class LastLoginfilter(FilterBase):
    class Meta:
        verbose_name = "Smart Filter: Last Login Check"
        verbose_name_plural = verbose_name

    days_since_login = models.IntegerField(
        default=30,
        help_text="Days since last login of any Character in the accounts before failing the filter.")

    no_data_pass = models.BooleanField(
        default=False,
        blank=True,
        help_text="If there is no data (No Valid Corp Token) for a characters account then should this filter automatically pass.")

    main_corp_only = models.BooleanField(
        default=False,
        blank=True,
        help_text="Only check characters in main's corporation.")

    def process_filter(self, user: User):
        try:
            check = self.audit_filter([user])

            if check[user.id]['check']:
                return True
            else:
                return False
        except Exception as e:
            logger.error(e, exc_info=1)
            return False

    def audit_filter(self, users):
        valid_logins = timezone.now(
        ) - datetime.timedelta(days=app_settings.CT_CHAR_MAX_INACTIVE_DAYS)
        co = CharacterOwnership.objects.filter(user__in=users, character__characteraudit__last_update_login__gte=valid_logins).select_related(
            "character__characteraudit", "character", "user__profile__main_character")

        chars = {}
        for c in co:
            if self.main_corp_only:
                if c.character.corporation_id != c.user.profile.main_character.corporation_id:
                    continue  # Skip this character

            if c.user_id not in chars:
                chars[c.user_id] = []

            if c.character.characteraudit.last_known_login:  # Login as Logoff in not always accurate
                chars[c.user_id].append(
                    # Login as Logoff in not always accurate
                    c.character.characteraudit.last_known_login)

        output = defaultdict(
            lambda: {"message": "No Data", "check": self.no_data_pass})
        for u in users:
            c = chars.get(u.id, False)
            if c is not False and len(c):
                max_date = max(c)
                string_date = max_date.strftime("%Y/%m/%d")
                days_since = (timezone.now() - max_date).days
                output[u.id] = {"message": f"{string_date} - {days_since} Days Ago",
                                "check": False if days_since > self.days_since_login else True}
                continue
        return output


class HomeStationFilter(FilterBase):
    class Meta:
        verbose_name = "Smart Filter: Home Station (Death Clone)"
        verbose_name_plural = verbose_name

    evelocation = models.ManyToManyField(
        EveLocation,
        blank=True,
        help_text="Limit filter to specific Structures",
    )

    def filter_query(self, users):
        character_list = CharacterOwnership.objects.filter(
            user__in=users).select_related('character', "character__characteraudit")

        character_count = Clone.objects.filter(
            character__character__in=character_list.values_list('character'))

        output = []

        if self.evelocation.all().count() > 0:
            output.append(models.Q(location_name__in=self.evelocation.all()))
            output.append(models.Q(
                location_id__in=self.evelocation.all().values_list('location_id', flat=True)))

        if len(output) > 0:
            query = output.pop()
            for _q in output:
                query |= _q
            character_count = character_count.filter(query)
        return character_count

    def process_filter(self, user: User):
        try:
            co = self.filter_query([user])
            if co.count() > 0:
                return True
            else:
                return False
        except Exception as e:
            # logger.exception(e)
            return False

    def audit_filter(self, users):
        co = self.filter_query(users).values("character__character__character_ownership__user",
                                             "character__character__character_name", "location_id")
        chars = defaultdict(dict)
        for c in co:
            uid = c["character__character__character_ownership__user"]
            char_name = c["character__character__character_name"]
            # This might be able to be optimized during the query. But theres no FK to work with?
            clone_location = EveLocation.objects.get(
                location_id=c['location_id']).location_name
            if char_name not in chars[uid]:
                chars[uid][char_name] = []
            chars[uid][char_name].append(clone_location)

        output = defaultdict(lambda: {"message": "", "check": False})
        for u in users:
            if len(chars[u.id]) > 0:
                ship_count = 0
                out_message = []
                for char, char_items in chars[u.id].items():
                    ship_count += len(char_items)
                    out_message.append(
                        f"{char}: {', '.join(list(set(char_items)))}")
                    output[u.id] = {"message": "<br>".join(
                        out_message), "check": True}
            else:
                output[u.id] = {"message": "", "check": False}
        return output


class JumpCloneFilter(FilterBase):
    class Meta:
        verbose_name = "Smart Filter: Jump Clone"
        verbose_name_plural = verbose_name

    evelocation = models.ManyToManyField(
        EveLocation, blank=True, help_text="Limit filter to specific Structures")

    def filter_query(self, users):
        character_list = CharacterOwnership.objects.filter(
            user__in=users).select_related('character', "character__characteraudit")

        character_count = JumpClone.objects.filter(
            character__character__in=character_list.values_list('character'))

        output = []

        if self.evelocation.all().count() > 0:
            output.append(models.Q(location_name__in=self.evelocation.all()))
            output.append(models.Q(
                location_id__in=self.evelocation.all().values_list('location_id', flat=True)))

        if len(output) > 0:
            query = output.pop()
            for _q in output:
                query |= _q
            character_count = character_count.filter(query)
        return character_count

    def process_filter(self, user: User):
        try:
            co = self.filter_query([user])
            if co.count() > 0:
                return True
            else:
                return False
        except Exception as e:
            logger.exception(e)
            return False

    def audit_filter(self, users):
        co = self.filter_query(users).values("character__character__character_ownership__user",
                                             "character__character__character_name", "location_id")
        chars = defaultdict(dict)
        for c in co:
            uid = c["character__character__character_ownership__user"]
            char_name = c["character__character__character_name"]
            # This might be able to be optimized during the query. But theres no FK to work with?
            clone_location = EveLocation.objects.get(
                location_id=c['location_id']).location_name
            if char_name not in chars[uid]:
                chars[uid][char_name] = []
            chars[uid][char_name].append(clone_location)

        output = defaultdict(lambda: {"message": "", "check": False})
        for u in users:
            if len(chars[u.id]) > 0:
                ship_count = 0
                out_message = []
                for char, char_items in chars[u.id].items():
                    ship_count += len(char_items)
                    out_message.append(
                        f"{char}: {', '.join(list(set(char_items)))}")
                    output[u.id] = {"message": "<br>".join(
                        out_message), "check": True}
            else:
                output[u.id] = {"message": "", "check": False}
        return output


class CharacterAgeFilter(FilterBase):
    class Meta:
        verbose_name = "Smart Filter: Character Age Filter"
        verbose_name_plural = verbose_name

    min_age = models.IntegerField(default=30)

    reversed_logic = models.BooleanField(
        default=False,
        help_text="If set all user with age less than the min_max_age will pass the test."
    )

    def process_filter(self, user: User):
        logic = self.reversed_logic
        try:
            main_character = user.profile.main_character.characteraudit
            histories = CorporationHistory.objects.filter(
                character=main_character).order_by('start_date').first()

            days = timezone.now() - histories.start_date
            if days.days >= self.min_age:
                return not logic
            else:
                return logic
        except Exception as e:
            # logger.error(e, exc_info=1)
            return False

    def audit_filter(self, users):
        logic = self.reversed_logic
        co = users.annotate(
            min_timestamp=Min(
                'profile__main_character__characteraudit__corporationhistory__start_date')
        ).values("id", "min_timestamp")

        chars = defaultdict(lambda: {})
        for c in co:
            if c['min_timestamp']:
                days = timezone.now() - c['min_timestamp']
                days = days.days
            else:
                days = -1
            chars[c['id']] = {
                "start_date": c["min_timestamp"],
                "age": days
            }

        output = defaultdict(lambda: {"message": "", "check": False})

        for c, char_corp_data in chars.items():
            start_date_localized = localize(char_corp_data["start_date"].date(
            )) if char_corp_data["start_date"] is not None else None
            end_date_localized = None

            if char_corp_data["age"] >= self.min_age:
                check = not logic
            else:
                if start_date_localized:
                    end_date_localized = localize(
                        (
                            char_corp_data["start_date"] +
                            datetime.timedelta(days=self.min_age)
                        ).date()
                    )

                check = logic
            if char_corp_data["age"] < 0:
                msg = "No Audit"
                check = False
            else:
                msg = (
                    ngettext_lazy(
                        singular=f'{char_corp_data["age"]:d} day old (Since: {start_date_localized})',
                        plural=f'{char_corp_data["age"]:d} days old (Since: {start_date_localized})',
                        number=char_corp_data["age"],
                    )
                    if not end_date_localized
                    else ngettext_lazy(
                        singular=f'{char_corp_data["age"]:d} day old (Until: {end_date_localized})',
                        plural=f'{char_corp_data["age"]:d} days old (Until: {end_date_localized})',
                        number=char_corp_data["age"],
                    )
                )
            output[c] = {"message": msg, "check": check}
        return output



import ast
import datetime
import json
from random import random

import yaml
from celery import chain as Chain, shared_task

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Count
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from esi.errors import TokenExpiredError
from esi.models import Token

from .. import app_settings, providers
from ..models import (
    CharacterAudit, CharacterBountyEvent, CharacterBountyStat,
    CharacterWalletJournalEntry, CorptoolsConfiguration, EveItemDogmaAttribute,
    EveItemType,
)
from ..task_helpers.char_tasks import (
    update_character_assets, update_character_clones,
    update_character_contacts, update_character_contract_items,
    update_character_contracts, update_character_industry_jobs,
    update_character_location, update_character_loyaltypoints,
    update_character_mail_headers, update_character_mining,
    update_character_notifications, update_character_order_history,
    update_character_orders, update_character_roles,
    update_character_skill_list, update_character_skill_queue,
    update_character_titles, update_character_transactions,
    update_character_wallet, update_corp_history,
)
from .locations import update_all_locations
from .updates import update_all_eve_names
from .utils import enqueue_next_task, esi_error_retry, no_fail_chain

TZ_STRING = "%Y-%m-%dT%H:%M:%SZ"


logger = get_extension_logger(__name__)


@shared_task(name="corptools.tasks.clear_all_etags")
def clear_all_etags():
    try:
        from django_redis import get_redis_connection
        _client = get_redis_connection("default")
    except (NotImplementedError, ModuleNotFoundError):
        from django.core.cache import caches
        default_cache = caches['default']
        _client = default_cache.get_master_client()

    keys = _client.keys(":?:etag-*")
    deleted = _client.delete(*keys)

    return f"Deleted {deleted} etag keys"


@shared_task(name="corptools.tasks.update_all_characters")
def update_all_characters():
    characters = CharacterAudit.objects.all().select_related('character')
    for char in characters:
        update_character.apply_async(args=[char.character.character_id])


@shared_task(bind=True, base=QueueOnce, name="corptools.tasks.update_subset_of_characters")
def update_subset_of_characters(self, subset=48, min_runs=15, force=False):
    amount_of_updates = max(
        CharacterAudit.objects.all().count() / subset, min_runs)
    characters = CharacterAudit.get_oldest_qs()[:amount_of_updates]
    char_ids = []
    for char in characters:
        char_ids.append(char.character.character_id)
        update_character.apply_async(args=[char.character.character_id], kwargs={
                                     "force_refresh": force})

    update_all_eve_names.apply_async(priority=7, kwargs={"chunk": 500})

    process_corp_histories.apply_async(priority=6)

    return f"Queued {len(characters)} Character Updates"


@shared_task(name="corptools.tasks.re_que_corp_histories")
def re_que_corp_histories():
    process_corp_histories.apply_async(priority=6)


@shared_task(bind=True, base=QueueOnce, name="corptools.tasks.process_corp_histories")
def process_corp_histories(self):
    cid = CharacterAudit.objects.all().order_by(
        'last_update_pub_data'
    ).first().character.character_id
    update_char_corp_history(cid)
    re_que_corp_histories.apply_async(countdown=2)
    return f"{(cid)} Character historys Updated"


@shared_task(name="corptools.tasks.check_account")
def check_account(character_id):
    char = EveCharacter.objects\
        .select_related('character_ownership',
                        'character_ownership__user__profile',
                        'character_ownership__user__profile__main_character', )\
        .get(character_id=int(character_id))\
        .character_ownership.user.profile.main_character

    linked_characters = char.character_ownership.user.character_ownerships.all(
    ).values_list('character__character_id', flat=True)
    for cid in linked_characters:
        update_character.apply_async(args=[cid], priority=6)


@shared_task(bind=True, base=QueueOnce, name="corptools.tasks.update_character")
def update_character(self, char_id, force_refresh=False):
    character = CharacterAudit.objects.filter(
        character__character_id=char_id).first()
    if character is None:
        token = Token.get_token(char_id, app_settings.get_character_scopes())
        if token:
            try:
                if token.valid_access_token():
                    character, created = CharacterAudit.objects.update_or_create(
                        character=EveCharacter.objects.get_character_by_id(token.character_id))
            except TokenExpiredError:
                return False
        else:
            logger.info(f"No Tokens for {char_id}")
            return False

    logger.info("Processing Updates for {}".format(
        character.character.character_name))

    skip_date = timezone.now() - datetime.timedelta(hours=2)  # hook into the settings

    que = []

    # TODO review this later
    if force_refresh:
        que.append(
            update_char_corp_history.si(
                character.character.character_id,
                force_refresh=force_refresh
            )
        )

    mindt = timezone.now() - datetime.timedelta(days=90)

    ct_conf = CorptoolsConfiguration.get_solo()

    if app_settings.CT_CHAR_ROLES_MODULE and not ct_conf.disable_update_roles:
        if (character.last_update_roles or mindt) <= skip_date or force_refresh:
            que.append(update_char_roles.si(
                character.character.character_id,
                force_refresh=force_refresh
            )
            )
        if (character.last_update_titles or mindt) <= skip_date or force_refresh:
            que.append(update_char_titles.si(
                character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_NOTIFICATIONS_MODULE and not ct_conf.disable_update_notif:
        if (character.last_update_notif or mindt) <= skip_date or force_refresh:
            que.append(update_char_notifications.si(
                character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_ASSETS_MODULE and not ct_conf.disable_update_assets:
        if (character.last_update_assets or mindt) <= skip_date or force_refresh:
            que.append(update_char_assets.si(
                character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_SKILLS_MODULE and not ct_conf.disable_update_skills:

        if (character.last_update_skills or mindt) <= skip_date or force_refresh:
            que.append(update_char_skill_list.si(
                character.character.character_id, force_refresh=force_refresh))

        if (character.last_update_skill_que or mindt) <= skip_date or force_refresh:
            que.append(update_char_skill_queue.si(
                character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_CLONES_MODULE and not ct_conf.disable_update_clones:
        if (character.last_update_clones or mindt) <= skip_date or force_refresh:
            que.append(update_char_clones.si(
                character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_CONTACTS_MODULE and not ct_conf.disable_update_contacts:
        if (character.last_update_contacts or mindt) <= skip_date or force_refresh:
            que.append(update_char_contacts.si(
                character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_MINING_MODULE and not ct_conf.disable_update_mining:
        if (character.last_update_mining or mindt) <= skip_date or force_refresh:
            que.append(update_char_mining_ledger.si(
                character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_WALLET_MODULE and not ct_conf.disable_update_wallet:
        if (character.last_update_wallet or mindt) <= skip_date or force_refresh:
            que.append(update_char_wallet.si(
                character.character.character_id, force_refresh=force_refresh))
            que.append(update_char_transactions.si(
                character.character.character_id, force_refresh=force_refresh))
        if (character.last_update_orders or mindt) <= skip_date or force_refresh:
            que.append(update_char_orders.si(
                character.character.character_id, force_refresh=force_refresh))
            que.append(update_char_order_history.si(
                character.character.character_id, force_refresh=force_refresh))
        if force_refresh or not app_settings.CT_CHAR_PAUSE_CONTRACTS:  # only on manual refreshes
            if (character.last_update_contracts or mindt) <= skip_date or force_refresh:
                que.append(update_char_contracts.si(
                    character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_LOCATIONS_MODULE and not ct_conf.disable_update_location:
        que.append(update_char_location.si(
            character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_MAIL_MODULE and not ct_conf.disable_update_mails:
        que.append(update_char_mail.si(
            character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_LOYALTYPOINTS_MODULE and not ct_conf.disable_update_loyaltypoints:
        que.append(update_char_loyaltypoints.si(
            character.character.character_id, force_refresh=force_refresh))

    if app_settings.CT_CHAR_INDUSTRY_MODULE and not ct_conf.disable_update_indy:
        que.append(update_char_industry_jobs.si(
            character.character.character_id, force_refresh=force_refresh))

    # We've updated this character, lets get all the missing locations.
    que.append(
        update_all_locations.s(
            character_filter=[character.character.character_id]
        )
    )

    if app_settings.CT_CHAR_SKILLS_MODULE and not ct_conf.disable_update_skills:
        # Must be last due to this being not a user level queue, Celery once will stall the queue here if characters on an account block.
        # TODO: make this better
        if (character.last_update_skills or mindt) <= skip_date or force_refresh:
            try:
                que.append(cache_user_skill_list.s(
                    character.character.character_ownership.user_id, force_refresh=force_refresh
                ))
            except ObjectDoesNotExist:
                pass

    # Spread out updates over 10 min to try be nice to ESI?
    delay = random() * app_settings.CT_TASK_SPREAD_DELAY
    if force_refresh:
        delay = 1  # If forced GO NOW!

    enqueue_next_task(que, delay=delay)

    logger.info(
        "Queued {} Updates for {} in {} seconds".format(
            len(que) + 1,
            character.character.character_name,
            delay
        )
    )


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_corp_history"
)
@no_fail_chain
@esi_error_retry
def update_char_corp_history(self, character_id, force_refresh=False, chain=[]):
    return update_corp_history(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_roles"
)
@no_fail_chain
@esi_error_retry
def update_char_roles(self, character_id, force_refresh=False, chain=[]):
    return update_character_roles(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_skill_list"
)
@no_fail_chain
@esi_error_retry
def update_char_skill_list(self, character_id, force_refresh=False, chain=[]):
    return update_character_skill_list(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_location"
)
@no_fail_chain
@esi_error_retry
def update_char_location(self, character_id, force_refresh=False, chain=[]):
    return update_character_location(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "keys": ["user_id"]
    },
    name="corptools.tasks.cache_user_skill_list"
)
@no_fail_chain
@esi_error_retry
def cache_user_skill_list(self, user_id, force_refresh=False, chain=[]):
    providers.skills.get_and_cache_user(
        user_id,
        force_rebuild=force_refresh
    )


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_industry_jobs"
)
@no_fail_chain
@esi_error_retry
def update_char_industry_jobs(self, character_id, force_refresh=False, chain=[]):
    return update_character_industry_jobs(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_skill_queue"
)
@no_fail_chain
@esi_error_retry
def update_char_skill_queue(self, character_id, force_refresh=False, chain=[]):
    return update_character_skill_queue(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_mining_ledger"
)
@no_fail_chain
@esi_error_retry
def update_char_mining_ledger(self, character_id, force_refresh=False, chain=[]):
    return update_character_mining(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_notifications"
)
@no_fail_chain
@esi_error_retry
def update_char_notifications(self, character_id, force_refresh=False, chain=[]):
    return update_character_notifications(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_assets"
)
@no_fail_chain
@esi_error_retry
def update_char_assets(self, character_id, force_refresh=False, chain=[]):
    return update_character_assets(
        character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_wallet"
)
@no_fail_chain
@esi_error_retry
def update_char_wallet(self, character_id, force_refresh=False, chain=[]):
    msg = update_character_wallet(character_id, force_refresh=force_refresh)

    for entry in CharacterWalletJournalEntry.objects.filter(
        character__character__character_id=character_id,
        ref_type="bounty_prizes",
        processed=False
    ):
        update_char_wallet_bounty_text.apply_async(
            args=[
                character_id,
                entry.entry_id
            ],
            priority=9
        )

    return msg


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_contacts"
)
@no_fail_chain
@esi_error_retry
def update_char_contacts(self, character_id, force_refresh=False, chain=[]):
    return update_character_contacts(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_orders"
)
@no_fail_chain
@esi_error_retry
def update_char_orders(self, character_id, force_refresh=False, chain=[]):
    return update_character_orders(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_transactions"
)
@no_fail_chain
@esi_error_retry
def update_char_transactions(self, character_id, force_refresh=False, chain=[]):
    return update_character_transactions(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_titles"
)
@no_fail_chain
@esi_error_retry
def update_char_titles(self, character_id, force_refresh=False, chain=[]):
    return update_character_titles(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_mail"
)
@no_fail_chain
@esi_error_retry
def update_char_mail(self, character_id, force_refresh=False, chain=[]):
    update_character_mail_headers(
        character_id, force_refresh=force_refresh)
    return "Completed mail pre-fetch for: %s" % str(character_id)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": True,
        "keys": ["character_id", "contract_id"]
    },
    name="corptools.tasks.update_char_contract_items"
)
def update_char_contract_items(self, character_id, contract_id, force_refresh=False):
    return update_character_contract_items(
        character_id, contract_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": True,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_contracts"
)
@no_fail_chain
@esi_error_retry
def update_char_contracts(self, character_id, force_refresh=False, chain=[]):
    _, ids = update_character_contracts(
        character_id, force_refresh=force_refresh)

    _chain = []
    for id in ids:
        _chain.append(update_char_contract_items.si(
            character_id, id, force_refresh=force_refresh))
    Chain(_chain).apply_async(priority=8)

    return "Completed Contracts for: %s" % str(character_id)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_order_history"
)
@no_fail_chain
@esi_error_retry
def update_char_order_history(self, character_id, force_refresh=False, chain=[]):
    return update_character_order_history(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_clones"
)
@no_fail_chain
@esi_error_retry
def update_char_clones(self, character_id, force_refresh=False, chain=[]):
    return update_character_clones(
        character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "graceful": False,
        "keys": ["character_id"]
    },
    name="corptools.tasks.update_char_loyaltypoints"
)
@no_fail_chain
@esi_error_retry
def update_char_loyaltypoints(self, character_id, force_refresh=False, chain=[]):
    return update_character_loyaltypoints(
        character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={
        "keys": ["character_id", "entry_id"]
    },
    name="corptools.tasks.translate_char_wallet_bounty_text"
)
def update_char_wallet_bounty_text(self, character_id, entry_id, force_refresh=False):
    logger.info(f"Translating {character_id} - {entry_id}")
    try:
        entry = CharacterWalletJournalEntry.objects.get(
            character__character__character_id=character_id,
            entry_id=entry_id,
            processed=False,
            ref_type="bounty_prizes"
        )
    except CharacterWalletJournalEntry.DoesNotExist:
        return "Unable to find {character_id} - {entry_id}"
    except CharacterWalletJournalEntry.MultipleObjectsReturned:
        clear_dupes(character_id, entry_id)
        self.retry()

    bounties = []

    logger.debug(f"Found {entry.reason}")

    for bty in entry.reason.split(","):
        b = bty.split(":")
        ship_type, _ = EveItemType.objects.get_or_create_from_esi(
            b[0].strip()
        )
        ship_dogma = EveItemDogmaAttribute.objects.filter(
            eve_type_id=ship_type.type_id,
            attribute_id=481
        ).first()
        bounties.append(
            {
                "msg": f"{b[1].strip()}x {ship_type.name} @ {ship_dogma.value:,.0f} ISK",
                "type_id": ship_type.type_id,
                "qty": b[1].strip()
            }
        )

    msg = "\n".join([b["msg"] for b in bounties])

    event = CharacterBountyEvent.objects.create(
        entry=entry,
        message=msg
    )
    for b in bounties:
        CharacterBountyStat.objects.create(
            event=event,
            type_name_id=b["type_id"],
            qty=b["qty"]
        )
    entry.processed = True
    entry.save()

    logger.info(f"Completed {character_id} - {entry_id}\n{event.message}")


def clear_dupes(character_id, entry_id):
    # dupe_ids = CharacterWalletJournalEntry.objects.filter(
    #     ref_type="bounty_prizes",
    #     character__character__character_id=character_id,
    #     entry_id=entry_id
    # ).values(
    #     'entry_id'
    # ).annotate(
    #     count=Count('entry_id')
    # ).filter(count__gt=1)

    logger.warning(f"Cleaning up `{entry_id}` duplicates in wallet.")

    ls = list(
        CharacterWalletJournalEntry.objects.filter(
            ref_type="bounty_prizes",
            character__character__character_id=character_id,
            entry_id=entry_id
        )
    )
    ls.pop()
    logger.warning(f"Found `{len(ls)}` duplicates.")
    for l in ls:
        l.delete()



from ninja import NinjaAPI

from django.db.models import Q

from allianceauth.services.hooks import get_extension_logger

from corptools import models
from corptools.api import schema

logger = get_extension_logger(__name__)


class AssetPingApiEndpoints:

    tags = ["Pinger"]

    def __init__(self, api: NinjaAPI):
        def build_ping_list(systems, structures, ignore_groups, message, filter_charges=False, ships_only=False, capitals_only=False):
            pingers = {}
            ammo_exclusions_cat = [8]
            filter_charges = True
            assets = models.CharacterAsset.objects.filter(
                Q(location_name_id__in=systems + structures) | Q(location_name__system_id__in=systems + structures)
            ).exclude(
                type_name__group_id__in=ignore_groups
            ).select_related(
                'type_name',
                'character',
                'character__character',
                'character__character__character_ownership',
                'character__character__character_ownership__user',
                'character__character__character_ownership__user__discord',
                'character__character__character_ownership__user__profile__main_character',
                'location_name'
            ).order_by("-type_name__volume"
                       )

            if filter_charges:
                assets = assets.exclude(
                    type_name__group__category_id__in=ammo_exclusions_cat)

            if ships_only:
                assets = assets.filter(
                    type_name__group__category_id__in=[6]
                )

            if capitals_only:
                assets = assets.filter(
                    type_name__group_id__in=[
                        30, 485, 513, 547, 659, 883, 902, 1538]
                )

            for a in assets:
                try:
                    uid = a.character.character.character_ownership.user.discord.uid
                    char = a.character.character.character_name
                    main = a.character.character.character_ownership.user.profile.main_character.character_name
                    if uid not in pingers:
                        pingers[uid] = {"c": set(), "a": list(),
                                        "s": set(), "m": main}
                    if char not in pingers[uid]:
                        pingers[uid]["c"].add(char)
                    if a.type_name.name not in pingers[uid]["a"]:
                        pingers[uid]["a"].append(a.type_name.name)
                    pingers[uid]["s"].add(a.location_name.location_name)
                except Exception:
                    pass

            return pingers

        @api.post(
            "pingbot/assets/send",
            response={200: schema.Message, 403: str},
            tags=self.tags
        )
        def post_send_pings_assets(request, message: str, systems: str = "", structures: str = "", ignore_groups: str = "", filter_charges: bool = False, ships_only: bool = False, capitals_only: bool = False):
            if not request.user.is_superuser:
                return 403, "Hard no pall!"

            from aadiscordbot.tasks import send_message
            from discord import Embed

            systems = systems.split(",") if len(systems) else []
            structures = structures.split(",") if len(structures) else []
            ignore_groups = ignore_groups.split(
                ",") if len(ignore_groups) else []
            pingers = build_ping_list(
                systems, structures, ignore_groups, message, filter_charges, ships_only, capitals_only)

            for id, chars in pingers.items():
                embed = Embed(title="Asset Alert!")
                embed.description = message.replace("\\n", "\n")
                _ = embed.add_field(name="Characters",
                                    value=", ".join(list(chars['c'])),
                                    inline=False)
                _ = embed.add_field(name="Structures",
                                    value=", ".join(list(chars['s'])),
                                    inline=False)
                _ = embed.add_field(name="Assets",
                                    value=", ".join(list(chars['a'])[:20]),
                                    inline=False)
                send_message(user_id=id, embed=embed)

            return 200, {"message": "Pings Sent!"}

        @api.post(
            "pingbot/assets/counts",
            response={200: schema.PingStats, 403: str},
            tags=self.tags
        )
        def post_test_pings_assets(request, systems: str = "", structures: str = "", ignore_groups: str = "", filter_charges: bool = False, ships_only: bool = False, capitals_only: bool = False):
            if not request.user.is_superuser:
                return 403, "Hard no pall!"

            systems = systems.split(",") if len(systems) else []
            structures = structures.split(",") if len(structures) else []
            ignore_groups = ignore_groups.split(
                ",") if len(ignore_groups) else []
            pingers = build_ping_list(
                systems, structures, ignore_groups, "", filter_charges, ships_only, capitals_only)

            locations = set()

            for id, chars in pingers.items():
                locations.update(list(chars['s']))

            locations = list(locations)
            locations.sort()

            return 200, {"members": len(pingers), "structures": locations}



# Generated by Django 3.2.8 on 2022-01-09 11:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0067_remove_notification_notification_text'),
    ]

    operations = [
        migrations.RenameField(
            model_name='notification',
            old_name='note_text',
            new_name='notification_text',
        ),
    ]



# Generated by Django 4.2.16 on 2025-02-11 11:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0112_alter_corptoolsconfiguration_holding_corps'),
    ]

    operations = [
        migrations.AddField(
            model_name='timeincorpfilter',
            name='reversed_logic',
            field=models.BooleanField(
                default=False, help_text='If set all members less than the days in corp will pass the test.'),
        ),
    ]



from typing import TYPE_CHECKING

from django.db import models

from .audits import CharacterAudit, CorporationAudit, EveLocation
from .eve_models import EveItemType, EveName

if TYPE_CHECKING:
    from esi.stubs import (
        CharactersCharacterIdContractsContractIdItemsGetItem,
        CharactersCharacterIdContractsGetItem,
    )


class Contract(models.Model):
    id = models.CharField(max_length=50, primary_key=True)

    contract_id = models.IntegerField()

    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)
    acceptor_id = models.IntegerField()
    acceptor_name = models.ForeignKey(
        EveName, on_delete=models.CASCADE, related_name="contractacceptor")
    assignee_id = models.IntegerField()
    assignee_name = models.ForeignKey(
        EveName, on_delete=models.CASCADE, related_name="contractassignee")
    issuer_id = models.IntegerField()
    issuer_name = models.ForeignKey(
        EveName, on_delete=models.CASCADE, related_name="contractissuer")
    issuer_corporation_id = models.IntegerField()
    issuer_corporation_name = models.ForeignKey(
        EveName, on_delete=models.CASCADE, related_name="contractissuercorporation")
    days_to_complete = models.IntegerField()

    collateral = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, blank=True)
    buyout = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, blank=True)
    price = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, blank=True)
    reward = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, blank=True)

    volume = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, blank=True)

    days_to_complete = models.IntegerField(null=True, blank=True)

    start_location_id = models.BigIntegerField(null=True, blank=True)
    start_location_name = models.ForeignKey(
        EveLocation, on_delete=models.SET_NULL, null=True, default=None, related_name="contractfrom")

    end_location_id = models.BigIntegerField(null=True, blank=True)
    end_location_name = models.ForeignKey(
        EveLocation, on_delete=models.SET_NULL, null=True, default=None, related_name="contractto")

    for_corporation = models.BooleanField()

    date_accepted = models.DateTimeField(null=True, blank=True)
    date_completed = models.DateTimeField(null=True, blank=True)

    date_expired = models.DateTimeField()
    date_issued = models.DateTimeField()

    status = models.CharField(max_length=25)
    contract_type = models.CharField(max_length=20)
    availability = models.CharField(max_length=15)

    title = models.CharField(max_length=256)

    @staticmethod
    def build_pk(char, cont):
        return f"{char}{cont}"

    class Meta:
        unique_together = (("character", "contract_id"),)

    @classmethod
    def from_esi_model(cls, character: CharacterAudit, esi_model: "CharactersCharacterIdContractsGetItem"):
        return cls(
            id=cls.build_pk(character.id, esi_model.contract_id),
            character=character,
            assignee_id=esi_model.assignee_id,
            assignee_name_id=esi_model.assignee_id,
            acceptor_id=esi_model.acceptor_id,
            acceptor_name_id=esi_model.acceptor_id,
            contract_id=esi_model.contract_id,
            availability=esi_model.availability,
            buyout=esi_model.buyout,
            collateral=esi_model.collateral,
            date_accepted=esi_model.date_accepted,
            date_completed=esi_model.date_completed,
            date_expired=esi_model.date_expired,
            date_issued=esi_model.date_issued,
            days_to_complete=esi_model.days_to_complete,
            end_location_id=esi_model.end_location_id,
            for_corporation=esi_model.for_corporation,
            issuer_corporation_name_id=esi_model.issuer_corporation_id,
            issuer_name_id=esi_model.issuer_id,
            price=esi_model.price,
            reward=esi_model.reward,
            start_location_id=esi_model.start_location_id,
            status=esi_model.status,
            title=esi_model.title,
            contract_type=esi_model.type,
            volume=esi_model.volume
        )


class ContractItem(models.Model):
    contract = models.ForeignKey(Contract, on_delete=models.CASCADE)
    is_included = models.BooleanField()
    is_singleton = models.BooleanField()
    quantity = models.IntegerField()
    raw_quantity = models.IntegerField(null=True, blank=True)
    record_id = models.BigIntegerField()
    type_name = models.ForeignKey(EveItemType, on_delete=models.CASCADE)

    class Meta:
        unique_together = (("record_id", "contract"),)

    @classmethod
    def from_esi_model(cls, contract: Contract, esi_model: "CharactersCharacterIdContractsContractIdItemsGetItem"):
        return cls(
            contract=contract,
            is_included=esi_model.is_included,
            is_singleton=esi_model.is_singleton,
            quantity=esi_model.quantity,
            raw_quantity=esi_model.raw_quantity,
            record_id=esi_model.record_id,
            type_name_id=esi_model.type_id,
        )


class CorporateContract(models.Model):
    id = models.CharField(max_length=50, primary_key=True)

    contract_id = models.BigIntegerField()

    corporation = models.ForeignKey(CorporationAudit, on_delete=models.CASCADE)
    acceptor_id = models.IntegerField()
    acceptor_name = models.ForeignKey(
        EveName, on_delete=models.CASCADE, related_name="corporationcontractacceptor")
    assignee_id = models.IntegerField()
    assignee_name = models.ForeignKey(
        EveName, on_delete=models.CASCADE, related_name="corporationcontractassignee")
    issuer_id = models.IntegerField()
    issuer_name = models.ForeignKey(
        EveName, on_delete=models.CASCADE, related_name="corporationcontractissuer")
    issuer_corporation_id = models.IntegerField()
    issuer_corporation_name = models.ForeignKey(
        EveName, on_delete=models.CASCADE, related_name="corporationcontractissuercorporation")
    days_to_complete = models.IntegerField()

    collateral = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, blank=True)
    buyout = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, blank=True)
    price = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, blank=True)
    reward = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, blank=True)

    volume = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, blank=True)

    days_to_complete = models.IntegerField(null=True, blank=True)

    start_location_id = models.BigIntegerField(null=True, blank=True)
    start_location_name = models.ForeignKey(
        EveLocation, on_delete=models.SET_NULL, null=True, default=None, related_name="corporationcontractfrom")

    end_location_id = models.BigIntegerField(null=True, blank=True)
    end_location_name = models.ForeignKey(
        EveLocation, on_delete=models.SET_NULL, null=True, default=None, related_name="corporationcontractto")

    for_corporation = models.BooleanField()

    date_accepted = models.DateTimeField(null=True, blank=True)
    date_completed = models.DateTimeField(null=True, blank=True)

    date_expired = models.DateTimeField()
    date_issued = models.DateTimeField()

    status = models.CharField(max_length=25)
    contract_type = models.CharField(max_length=20)
    availability = models.CharField(max_length=15)

    title = models.CharField(max_length=256)

    @staticmethod
    def build_pk(corp, cont):
        return f"{corp}{cont}"

    class Meta:
        unique_together = (("corporation", "contract_id"),)


class CorporateContractItem(models.Model):
    contract = models.ForeignKey(CorporateContract, on_delete=models.CASCADE)
    is_included = models.BooleanField()
    is_singleton = models.BooleanField()
    quantity = models.IntegerField()
    raw_quantity = models.IntegerField(null=True, blank=True)
    record_id = models.BigIntegerField()
    type_name = models.ForeignKey(EveItemType, on_delete=models.CASCADE)

    class Meta:
        unique_together = (("record_id", "contract"),)



# Generated by Django 2.2.12 on 2020-08-10 13:33

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0021_auto_20200804_0941'),
    ]

    operations = [
        migrations.CreateModel(
            name='SkillList',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('Name', models.CharField(default=None, max_length=500, null=True)),
                ('skill_list', models.TextField(default='', null=True)),
                ('eft', models.TextField(default='', null=True)),
            ],
        ),
        migrations.CreateModel(
            name='MapSystemGate',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('from_solar_system', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                                        related_name='from_system', to='corptools.MapSystem')),
                ('to_solar_system', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                                      related_name='to_system', to='corptools.MapSystem')),
            ],
        ),
    ]



from typing import TYPE_CHECKING

from django.db import models

from .audits import CharacterAudit
from .eve_models import EveItemType, MapSystem

if TYPE_CHECKING:
    from esi.stubs import CharactersCharacterIdMiningGetItem


class CharacterMiningLedger(models.Model):
    id = models.CharField(max_length=50, primary_key=True)
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)
    date = models.DateField()
    type_name = models.ForeignKey(EveItemType, on_delete=models.CASCADE)
    system = models.ForeignKey(MapSystem, on_delete=models.CASCADE)
    quantity = models.IntegerField()

    @staticmethod
    def create_primary_key(character_id, mining_record: "CharactersCharacterIdMiningGetItem"):
        """
            TODO investigate something else...
        """
        return f"{mining_record.date.strftime('%Y%m%d')}-{mining_record.type_id}-{character_id}-{mining_record.solar_system_id}"
